#include "valence.h"

#include <stdbool.h>
#include <stdio.h>
#include <stdint.h>
#include <time.h>

#include <velib/canhw/canhw.h>
#include <velib/velib_inc_J1939.h>
#include <velib/platform/console.h>
#include <velib/types/ve_dbus_item.h>
#include <velib/types/ve_item.h>
#include <velib/types/ve_item_def.h>
#include <velib/types/ve_values.h>
#include <velib/types/variant_print.h>
#include <velib/canhw/canhw_types.h>
#include <velib/utils/ve_timer.h>
#include <velib/types/variant.h>
#include <velib/vecan/products.h>

#include "../velib/inc/velib/canhw/canhw_types.h"

#define PROCESS_NAME "dbus_valence"
#define MODULE "VALENCE"
#define BMS_TIMEOUT   5*20 /* *50ms = 5 sec */

static char const *SETTINGS_SERVICE_NAME = "com.victronenergy.settings";

static char const *BMS_PRODUCT_NAME = "Valence BMS";
static char const *BATTERY_PRODUCT_NAME = "Valence Battery";

/**
 * When announced maximum regenerative current goes below this value the
 * allow to charge flag is set to zero.
 */
#define MAX_REGENERATIVE_CURRENT_THRESHOLD 1

/**
 * When announced maximum discharge current is below this value the allow to
 * discharge flag is set to zero.
 */
#define MAX_DISCHARGE_CURRENT_THRESHOLD 1

/** How often to push values to dbus in ms? */
#define UPDATE_INTERVAL 1000

enum BatteryMode { MODE_STANDBY = 0x00, MODE_CHARGE = 0x01, MODE_DRIVE = 0x02 };

static const un8 MODULE_TYPE_MAX_VAL = 1;
enum ModuleType { TYPE_U24 = 0, TYPE_U27 = 1 };

static const int MESSAGE_ID_VALENCE_BATTERY_STATUS = 0x0c0;
static const int MESSAGE_ID_VALENCE_BATTERY_INFO = 0x0c1;
static const int MESSAGE_ID_VALENCE_VMU_TRACE = 0x0c4;
static const int MESSAGE_ID_VALENCE_HARDWARE_FIRMWARE_REV = 0x180;

typedef struct {
  struct VeDbus *conn;
  VeItem root;

  VeItem Connected;
  VeItem ProductName;
  VeItem ProductId;
  VeItem DeviceInstance;
  VeItem Connection;
  VeItem N2KUniqueNumber;
  VeItem FirmwareVersion;
  VeItem HardwareVersion;
  VeItem Serial;
  VeItem ProcessName;
  VeItem ProcessVersion;

  VeItem NrOfBatteries;
  VeItem BatteriesParallel;
  VeItem BatteriesSeries;
  VeItem NrOfCellsPerBattery;
  VeItem Balancing;
  VeItem State;
  VeItem Error;
  VeItem SystemSwitch;
  VeItem MinCellVoltage;
  VeItem MaxCellVoltage;
  VeItem AllowToCharge;
  VeItem AllowToDischarge;
} BMSDbus;

typedef struct {
  struct VeDbus *conn;
  VeItem root;

  VeItem Connected;
  VeItem ProductName;
  VeItem ProductId;
  VeItem DeviceInstance;
  VeItem Connection;
  VeItem N2KUniqueNumber;
  VeItem FirmwareVersion;
  VeItem HardwareVersion;
  VeItem Serial;
  VeItem ProcessName;
  VeItem ProcessVersion;

  VeItem Voltage;
  VeItem Current;
  VeItem Power;
  VeItem Temperature;
  VeItem ConsumedAmphours;
  VeItem Soc;
  VeItem TimeToGo;
} BatteryDbus;

typedef struct {
  uint8_t firmwareVersion;
  bool hasErrors;
  bool balancing;
  enum ModuleType moduleType;
  uint8_t nrModules;
  uint8_t nrOfCellsPerBattery;
  uint8_t parallel;
  uint8_t series;
  float minCellVoltage;
  float maxCellVoltage;
  unsigned int maxRegenCurrent;
  unsigned int maxDischargeCurrent;

  BMSDbus dbus;
} BMS;

typedef struct {
  enum BatteryMode mode;
  float capacity;
  float temperature;
  float voltage;
  float current;
  float power;
  float soc;
  float consumedAmphours;
  float timeToGo;

  BatteryDbus dbus;
} Battery;

static bool initBMS(BMS *bms);
static bool initBattery(Battery *bat);

/**
* Compute battery parameters and total capacity of the system.
*
* @param moduleType Type of the modules.
* @param numModules Total number of battery modules in the system.
*/
static void updateBatteryConfiguration(enum ModuleType moduleType, uint8_t numModules, BMS *bms, Battery *bat);
static void updateDbusBMS(BMS *bms);
static void updateDbusBattery(Battery *bat);

static Battery bat;
static BMS bms;
static struct VeDbus *settingsConn;
static VeItem settingsRoot;
static un16 timeout;
/** When the values were pushed to dbus. */
static un16 lastUpdate;

static VeVariantUnitFmt none = {	0,	""		};
static VeVariantUnitFmt current = {	0,	"A"		};
static VeVariantUnitFmt voltage = {	0,	"V"		};
static VeVariantUnitFmt cellVoltage = {	2,	"V"		};
static VeVariantUnitFmt temperature = {	0,	"C"		};
static VeVariantUnitFmt capacity = {	0,	"Ah"		};
static VeVariantUnitFmt power = {	0,	"W"		};
static VeVariantUnitFmt percentage = {	0,	"%"		};

static void batteryModuleTypeChanged(VeItem *type)
{
  VeVariant var;

  veItemValue(type, &var);
  un8 typeCode = var.value.UN8;
  if (typeCode > MODULE_TYPE_MAX_VAL)
  {
    printf("Warning! Unknown battery type %u. Defaulting to type U27.\n", typeCode);
    typeCode = TYPE_U27;
  }
  enum ModuleType moduleType = (enum ModuleType)typeCode;
  printf("Battery's module type is now %s.\n", typeCode == TYPE_U24 ? "U24" : "U27");
  updateBatteryConfiguration(moduleType, bms.nrModules, &bms, &bat); // FIXME thread safety
}

static bool listenModuleTypeSetting(VeItem *root)
{
  VeItem *consumer = veItemGetOrCreateUid(root, SETTINGS_SERVICE_NAME);
  if (!veDbusAddRemoteService(SETTINGS_SERVICE_NAME, consumer, veTrue)) {
    printf("veDbusAddRemoteService failed\n");
    return false;
  }
  VeItem *value = veItemGetOrCreateUid(consumer, "Settings/Valence/ModuleType");
  veItemSetChanged(value, batteryModuleTypeChanged);
  batteryModuleTypeChanged(value);
  return true;
}

void valenceWait(void)
{
  printf("valenceWait\n");

  VeRawCanMsg msg;
  un8 found;
  struct timespec sleeptime;
  sleeptime.tv_sec = 0;
  sleeptime.tv_nsec = 100000000L; /* 100 ms */

  /* set when Valence BMS found */
  found = 0;

  /* Loop until we see Valence BATTERY STATUS message */
  while (!found)
  {
    while (veCanRead(&msg))
    {
      if (msg.canId == MESSAGE_ID_VALENCE_BATTERY_STATUS)
      {
        printf("Valence BMS Instance 0 found\n");
        timeout = BMS_TIMEOUT;
        found = 1;
      }
    }
    nanosleep(&sleeptime, NULL);
  }

}

/* Connect to the dbus and publish some values */
void valenceInit(void)
{
  printf("valenceInit\n");

  /* The ccgx uses the system dbus, a good candidate for an argument */
#if defined(TARGET_pc)
  settingsConn = veDbusConnect(DBUS_BUS_SESSION);
  bms.dbus.conn = veDbusConnect(DBUS_BUS_SESSION);
  bat.dbus.conn = veDbusConnect(DBUS_BUS_SESSION);
#else
  settingsConn = veDbusConnect(DBUS_BUS_SYSTEM);
  bms.dbus.conn = veDbusConnect(DBUS_BUS_SYSTEM);
  bat.dbus.conn = veDbusConnect(DBUS_BUS_SYSTEM);
#endif

  if ((bms.dbus.conn == NULL) || (bat.dbus.conn == NULL))
  {
    if (bms.dbus.conn == NULL)
      printf("dbus_service: no D-Bus connection for BMS\n");
    else
      veDbusDisconnect(bms.dbus.conn);

    if (bat.dbus.conn == NULL)
      printf("dbus_service: no D-Bus connection for battery\n");
    else
      veDbusDisconnect(bat.dbus.conn);

    pltExit(10);
  }

  if (!initBMS(&bms))
    pltExit(11);
  if (!initBattery(&bat))
    pltExit(12);

  veDbusSetListeningDbus(settingsConn);
  if (!listenModuleTypeSetting(&settingsRoot))
  {
    printf("Could not get module type setting. Defaulting to U27.");
    updateBatteryConfiguration(TYPE_U27, bms.nrModules, &bms, &bat);
  }
}

static bool initBMS(BMS *bms)
{
  VeVariant variant;

  veDbusItemInit(bms->dbus.conn, &bms->dbus.root);
  if (!veDbusChangeName(bms->dbus.conn, "com.victronenergy.bms.valence0"))
  {
    printf("dbus_service: registering name for BMS failed\n");
    return false;
  }

  VeItem *root = &bms->dbus.root;

  veItemAddChildByUid(root, "ProductName", &bms->dbus.ProductName);
  veItemSetFmt(&bms->dbus.ProductName, veVariantFmt, &none);
  veItemAddChildByUid(root, "ProductId", &bms->dbus.ProductId);
  veItemSetFmt(&bms->dbus.ProductId, veVariantFmt, &none);
  veItemAddChildByUid(root, "FirmwareVersion", &bms->dbus.FirmwareVersion);
  veItemSetFmt(&bms->dbus.FirmwareVersion, veVariantFmt, &none);
  veItemAddChildByUid(root, "HardwareVersion", &bms->dbus.HardwareVersion);
  veItemSetFmt(&bms->dbus.HardwareVersion, veVariantFmt, &none);
  veItemAddChildByUid(root, "Serial", &bms->dbus.Serial);
  veItemSetFmt(&bms->dbus.Serial, veVariantFmt, &none);
  veItemAddChildByUid(root, "DeviceInstance", &bms->dbus.DeviceInstance);
  veItemSetFmt(&bms->dbus.DeviceInstance, veVariantFmt, &none);
  veItemAddChildByUid(root, "State", &bms->dbus.State);
  veItemSetFmt(&bms->dbus.State, veVariantFmt, &none);
  veItemAddChildByUid(root, "N2KUniqueNumber", &bms->dbus.N2KUniqueNumber);
  veItemSetFmt(&bms->dbus.N2KUniqueNumber, veVariantFmt, &none);
  veItemAddChildByUid(root, "Connected", &bms->dbus.Connected);
  veItemSetFmt(&bms->dbus.Connected, veVariantFmt, &none);
  veItemAddChildByUid(root, "Mgmt/ProcessName", &bms->dbus.ProcessName);
  veItemSetFmt(&bms->dbus.ProcessName, veVariantFmt, &none);
  veItemAddChildByUid(root, "Mgmt/ProcessVersion", &bms->dbus.ProcessVersion);
  veItemSetFmt(&bms->dbus.ProcessVersion, veVariantFmt, &none);
  veItemAddChildByUid(root, "Mgmt/Connection", &bms->dbus.Connection);
  veItemSetFmt(&bms->dbus.Connection, veVariantFmt, &none);

  veItemAddChildByUid(root, "System/NrOfBatteries", &bms->dbus.NrOfBatteries);
  veItemSetFmt(&bms->dbus.NrOfBatteries, veVariantFmt, &none);
  veItemAddChildByUid(root, "System/BatteriesParallel", &bms->dbus.BatteriesParallel);
  veItemSetFmt(&bms->dbus.BatteriesParallel, veVariantFmt, &none);
  veItemAddChildByUid(root, "System/BatteriesSeries", &bms->dbus.BatteriesSeries);
  veItemSetFmt(&bms->dbus.BatteriesSeries, veVariantFmt, &none);
  veItemAddChildByUid(root, "System/NrOfCellsPerBattery",&bms->dbus.NrOfCellsPerBattery);
  veItemSetFmt(&bms->dbus.NrOfCellsPerBattery, veVariantFmt, &none);
  veItemAddChildByUid(root, "System/MinCellVoltage", &bms->dbus.MinCellVoltage);
  veItemSetFmt(&bms->dbus.MinCellVoltage, veVariantFmt, &cellVoltage);
  veItemAddChildByUid(root, "System/MaxCellVoltage", &bms->dbus.MaxCellVoltage);
  veItemSetFmt(&bms->dbus.MaxCellVoltage, veVariantFmt, &cellVoltage);
  veItemAddChildByUid(root, "Balancing", &bms->dbus.Balancing);
  veItemSetFmt(&bms->dbus.Balancing, veVariantFmt, &none);
  veItemAddChildByUid(root, "Io/AllowToCharge", &bms->dbus.AllowToCharge);
  veItemSetFmt(&bms->dbus.AllowToCharge, veVariantFmt, &none);
  veItemAddChildByUid(root, "Io/AllowToDischarge", &bms->dbus.AllowToDischarge);
  veItemSetFmt(&bms->dbus.AllowToDischarge, veVariantFmt, &none);

  // Send basic information once
  veItemOwnerSet(&bms->dbus.ProcessName, veVariantStr(&variant, PROCESS_NAME));
  veItemOwnerSet(&bms->dbus.ProcessVersion, veVariantStr(&variant, VERSION_STR));
  veItemOwnerSet(&bms->dbus.Connected, veVariantUn8(&variant, 1));
  veItemOwnerSet(&bms->dbus.ProductName, veVariantStr(&variant, BMS_PRODUCT_NAME));
  veItemOwnerSet(&bms->dbus.ProductId, veVariantUn32(&variant, VE_PROD_ID_VALENCE_BMS));
  veItemOwnerSet(&bms->dbus.DeviceInstance, veVariantUn8(&variant, 0));
  veItemOwnerSet(&bms->dbus.Connected, veVariantUn8(&variant, 1));
  veItemOwnerSet(&bms->dbus.DeviceInstance, veVariantUn8(&variant, 0));
  veItemOwnerSet(&bms->dbus.Serial, veVariantStr(&variant, ""));
  veItemOwnerSet(&bms->dbus.HardwareVersion, veVariantStr(&variant, ""));
  veItemOwnerSet(&bms->dbus.Connection, veVariantStr(&variant, "VE.Can"));
  veItemOwnerSet(&bms->dbus.State, veVariantUn32(&variant, 9));
  veItemOwnerSet(&bms->dbus.N2KUniqueNumber, veVariantUn32(&variant, 10));
  veItemOwnerSet(&bms->dbus.SystemSwitch, veVariantUn32(&variant, 1));

  updateDbusBMS(bms);
  return true;
}

static bool initBattery(Battery *bat)
{
  VeVariant variant;

  veDbusItemInit(bat->dbus.conn, &bat->dbus.root);
  if (!veDbusChangeName(bat->dbus.conn, "com.victronenergy.battery.valence0"))
  {
    printf("dbus_service: registering name for battery failed\n");
    return false;
  }

  VeItem *root = &bat->dbus.root;

  veItemAddChildByUid(root, "ProductName", &bat->dbus.ProductName);
  veItemSetFmt(&bat->dbus.ProductName, veVariantFmt, &none);
  veItemAddChildByUid(root, "ProductId", &bat->dbus.ProductId);
  veItemSetFmt(&bat->dbus.ProductId, veVariantFmt, &none);
  veItemAddChildByUid(root, "FirmwareVersion", &bat->dbus.FirmwareVersion);
  veItemSetFmt(&bat->dbus.FirmwareVersion, veVariantFmt, &none);
  veItemAddChildByUid(root, "HardwareVersion", &bat->dbus.HardwareVersion);
  veItemSetFmt(&bat->dbus.HardwareVersion, veVariantFmt, &none);
  veItemAddChildByUid(root, "Serial", &bat->dbus.Serial);
  veItemSetFmt(&bat->dbus.Serial, veVariantFmt, &none);
  veItemAddChildByUid(root, "DeviceInstance", &bat->dbus.DeviceInstance);
  veItemSetFmt(&bat->dbus.DeviceInstance, veVariantFmt, &none);
  veItemAddChildByUid(root, "N2KUniqueNumber", &bat->dbus.N2KUniqueNumber);
  veItemSetFmt(&bat->dbus.N2KUniqueNumber, veVariantFmt, &none);
  veItemAddChildByUid(root, "Connected", &bat->dbus.Connected);
  veItemSetFmt(&bat->dbus.Connected, veVariantFmt, &none);
  veItemAddChildByUid(root, "Mgmt/ProcessName", &bat->dbus.ProcessName);
  veItemSetFmt(&bat->dbus.ProcessName, veVariantFmt, &none);
  veItemAddChildByUid(root, "Mgmt/ProcessVersion", &bat->dbus.ProcessVersion);
  veItemSetFmt(&bat->dbus.ProcessVersion, veVariantFmt, &none);
  veItemAddChildByUid(root, "Mgmt/Connection", &bat->dbus.Connection);
  veItemSetFmt(&bat->dbus.Connection, veVariantFmt, &none);

  veItemAddChildByUid(root, "Dc/0/Voltage", &bat->dbus.Voltage);
  veItemSetFmt(&bat->dbus.Voltage, veVariantFmt, &voltage);
  veItemAddChildByUid(root, "Dc/0/Current", &bat->dbus.Current);
  veItemSetFmt(&bat->dbus.Current, veVariantFmt, &current);
  veItemAddChildByUid(root, "Dc/0/Power", &bat->dbus.Power);
  veItemSetFmt(&bat->dbus.Power, veVariantFmt, &power);
  veItemAddChildByUid(root, "Dc/0/Temperature", &bat->dbus.Temperature);
  veItemSetFmt(&bat->dbus.Temperature, veVariantFmt, &temperature);
  veItemAddChildByUid(root, "ConsumedAmphours", &bat->dbus.ConsumedAmphours);
  veItemSetFmt(&bat->dbus.ConsumedAmphours, veVariantFmt, &capacity);
  veItemAddChildByUid(root, "Soc", &bat->dbus.Soc);
  veItemSetFmt(&bat->dbus.Soc, veVariantFmt, &percentage);
  veItemAddChildByUid(root, "TimeToGo", &bat->dbus.TimeToGo);
  veItemSetFmt(&bat->dbus.TimeToGo, veVariantFmt, &none);

  // Send basic information once
  veItemOwnerSet(&bat->dbus.FirmwareVersion, veVariantUn32(&variant, 0));

  veItemOwnerSet(&bat->dbus.ProcessName, veVariantStr(&variant, PROCESS_NAME));
  veItemOwnerSet(&bat->dbus.ProcessVersion, veVariantStr(&variant, VERSION_STR));
  veItemOwnerSet(&bat->dbus.Connected, veVariantUn8(&variant, 1));
  veItemOwnerSet(&bat->dbus.ProductName, veVariantStr(&variant, BATTERY_PRODUCT_NAME));
  veItemOwnerSet(&bat->dbus.ProductId, veVariantUn32(&variant, VE_PROD_ID_VALENCE_XP_BATTERY));
  veItemOwnerSet(&bat->dbus.DeviceInstance, veVariantUn8(&variant, 0));
  veItemOwnerSet(&bat->dbus.Connected, veVariantUn8(&variant, 1));
  veItemOwnerSet(&bat->dbus.DeviceInstance, veVariantUn8(&variant, 0));
  veItemOwnerSet(&bat->dbus.Serial, veVariantStr(&variant, ""));
  veItemOwnerSet(&bat->dbus.HardwareVersion, veVariantStr(&variant, ""));
  veItemOwnerSet(&bat->dbus.Connection, veVariantStr(&variant, "VE.Can"));
  veItemOwnerSet(&bat->dbus.N2KUniqueNumber, veVariantUn32(&variant, 11));

  updateDbusBattery(bat);
  return true;
}

/**
 * Return true if there are some error(s) present, false otherwise.
 */
static bool hasError(const VeRawCanMsg *u_bms_vmu_status)
{
  const un8 BYTE2_MASK = 0xBF; // ignore bit 6 (reserved)
  const un8 BYTE3_MASK = 0xFC; // ignore bits 0 and 1 (reserved)
  const un8 BYTE7_MASK = 0xBC; // ignore bits 0 and 1 (reserved) and bit 6 (sanity error)
  return (u_bms_vmu_status->mdata[2] & BYTE2_MASK) ||
      (u_bms_vmu_status->mdata[3] & BYTE3_MASK) ||
      (u_bms_vmu_status->mdata[7] & BYTE7_MASK);
}

void valenceUpdate(void)
{
  //printf("valenceUpdate\n");

  VeRawCanMsg msg;

  /* veCanRead returns false if there is no message to be read */
  while (veCanRead(&msg))
  {
    if ((msg.canId == MESSAGE_ID_VALENCE_BATTERY_STATUS) && (msg.length == 8))
    {
      bat.soc = (float)(unsigned short)msg.mdata[0];
      bat.mode = (enum BatteryMode)(msg.mdata[1] & 0x03);
      bms.balancing = msg.mdata[6] ? true : false;
      bms.hasErrors = hasError(&msg);
      uint8_t numModules = (uint8_t)(msg.mdata[5]);
      updateBatteryConfiguration(bms.moduleType, numModules, &bms, &bat);
      timeout = BMS_TIMEOUT;
    }
    else if ((msg.canId == MESSAGE_ID_VALENCE_BATTERY_INFO) && (msg.length == 8))
    {
      bat.voltage = (float)(unsigned short)(msg.mdata[0]) * 2;
      bat.current = (float)(msg.mdata[2] << 8) + msg.mdata[1] - 0x8000;
      bat.power = bat.voltage * bat.current;
      bat.consumedAmphours = bat.capacity * bat.soc / 100.0f - bat.capacity;

      if (bat.current < 0.0f)
      {
        bat.timeToGo = bat.capacity * bat.soc / 100.0f / -bat.current * 60.0f * 60.0f;
      }
      else
      {
        bat.timeToGo = 0.0f;
      }

      if (bat.mode == MODE_DRIVE)
      {
        bms.maxDischargeCurrent = ((unsigned int)(msg.mdata[4]) << 8) + msg.mdata[3];
        bms.maxRegenCurrent = ((unsigned int)(msg.mdata[7]) << 8) + msg.mdata[5];
      }
      else
      {
        // we know these only in drive mode
        bms.maxDischargeCurrent = 0;
        bms.maxRegenCurrent = 0;
      }
    }
    else if ((msg.canId == MESSAGE_ID_VALENCE_VMU_TRACE) && (msg.length == 8))
    {
      bat.temperature = (float)(msg.mdata[0]) - 40.0f; // Battery max temperature

      unsigned int voltage = (unsigned int)(msg.mdata[7]) << 8;
      voltage += (unsigned int)(msg.mdata[6]);
      bms.minCellVoltage = (float)voltage / 1000.0f;

      voltage = (unsigned int)(msg.mdata[5]) << 8;
      voltage += (unsigned int)(msg.mdata[4]);
      bms.maxCellVoltage = (float)voltage / 1000.0f;
    }
    else if ((msg.canId == MESSAGE_ID_VALENCE_HARDWARE_FIRMWARE_REV) && (msg.length == 8))
    {
      bms.firmwareVersion = msg.mdata[0];
    }

    if (veTick1ms(&lastUpdate, UPDATE_INTERVAL))
    {
      updateDbusBMS(&bms);
      updateDbusBattery(&bat);
    }
    // Print for Debugging
    //printf("Battery SOC: %.0f timetogo: %.1f voltage: %.1f current: %.1f temp: %0.f\n",
    //	     batterySoc,
    //	     batteryTimeToGo,
    //     batteryVoltage,
    //     batteryCurrent,
    //     batteryTemperature);


  }
}

void valenceTick(void)
{

  if (timeout)
  {
    timeout--;
    if (timeout == 0)
    {
      printf("Valence BMS timed out\n");
      pltExit(129);
    }

  }

}

static void updateBatteryConfiguration(enum ModuleType moduleType, uint8_t numModules, BMS *bms, Battery *bat)
{
  bms->nrOfCellsPerBattery = 4;
  bms->series = 4;
  bms->nrModules = numModules;
  bms->parallel = numModules / bms->series;
  bms->moduleType = moduleType;
  switch (moduleType)
  {
    case TYPE_U24:
      bat->capacity = bms->parallel * 110;
      break;
    case TYPE_U27:
      bat->capacity = bms->parallel * 138;
      break;
  }
}

static void updateDbusBMS(BMS *bms)
{
  VeVariant variant;

  uint8_t allowToDischarge = 0;
  if (bms->maxDischargeCurrent >= MAX_DISCHARGE_CURRENT_THRESHOLD)
    allowToDischarge = 1;
  veItemOwnerSet(&bms->dbus.AllowToDischarge, veVariantUn8(&variant, allowToDischarge));

  uint8_t allowToCharge = 0;
  if (bms->maxRegenCurrent >= MAX_REGENERATIVE_CURRENT_THRESHOLD)
    allowToCharge = 1;
  veItemOwnerSet(&bms->dbus.AllowToCharge, veVariantUn8(&variant, allowToCharge));

  veItemOwnerSet(&bms->dbus.FirmwareVersion, veVariantUn8(&variant, bms->firmwareVersion));
  veItemOwnerSet(&bms->dbus.Error, veVariantUn8(&variant, bms->hasErrors ? 1 : 0));
  veItemOwnerSet(&bms->dbus.Balancing, veVariantUn8(&variant, bms->balancing ? 1 : 0));
  veItemOwnerSet(&bms->dbus.NrOfBatteries, veVariantUn8(&variant, bms->nrModules));
  veItemOwnerSet(&bms->dbus.BatteriesParallel, veVariantUn8(&variant, bms->parallel));
  veItemOwnerSet(&bms->dbus.BatteriesSeries, veVariantUn8(&variant, bms->series));
  veItemOwnerSet(&bms->dbus.NrOfCellsPerBattery, veVariantUn8(&variant, bms->nrOfCellsPerBattery));
  veItemOwnerSet(&bms->dbus.MinCellVoltage, veVariantFloat(&variant, bms->minCellVoltage));
  veItemOwnerSet(&bms->dbus.MaxCellVoltage, veVariantFloat(&variant, bms->maxCellVoltage));
}

static void updateDbusBattery(Battery *bat)
{
  VeVariant variant;

  veItemOwnerSet(&bat->dbus.Voltage, veVariantFloat(&variant, bat->voltage));
  veItemOwnerSet(&bat->dbus.Current, veVariantFloat(&variant, bat->current));
  veItemOwnerSet(&bat->dbus.Power, veVariantFloat(&variant, bat->power));
  veItemOwnerSet(&bat->dbus.Soc, veVariantFloat(&variant, bat->soc));
  veItemOwnerSet(&bat->dbus.ConsumedAmphours, veVariantFloat(&variant, bat->consumedAmphours));
  veItemOwnerSet(&bat->dbus.TimeToGo, veVariantFloat(&variant, bat->timeToGo));
  veItemOwnerSet(&bat->dbus.Temperature, veVariantFloat(&variant, bat->temperature));
}
